import copy
from fjlc.utils.json_utils import from_json_file


class PriorPolarityLexicon:

    def __init__(self, polarity_lexicon):

        # Check if file name was given instead of ready polarity lexicon
        if type(polarity_lexicon) is str:
            self.polarity_lexicon = self.read_lexicon(polarity_lexicon)
        else:
            self.polarity_lexicon = polarity_lexicon

    def get_token_polarity(self, phrase):
        return self.polarity_lexicon[phrase]

    def has_token(self, word):
        return word in self.polarity_lexicon

    def get_subjective_words(self):
        return self.polarity_lexicon.keys()

    def get_lexicon(self):
        return copy.deepcopy(self.polarity_lexicon)

    @staticmethod
    def read_lexicon(file_name):
        return from_json_file(file_name)



import json

import fjlc.utils.file_utils as file_utils


def to_json(data, pretty):
    """
    Converts object to JSON formatted string with typeToken adapter
    :param data: A dictionary to convert to JSON string
    :param pretty: A boolean deciding whether or not to pretty format the JSON string
    :return: The JSON string
    """
    if pretty:
        return json.dumps(data, sort_keys=True, indent=4, separators=(',', ': '))
    return json.dumps(data)


def to_json_file(file, data, pretty):
    """
    Writes object instance in JSON formatted String to file
    
    :param file: File to write JSON string ot
    :param data: Object to convert to JSON 
    :param pretty: Use pretty formatting or not
    """
    json_string = to_json(data, pretty)
    file_utils.write_to_file(file, json_string)


def from_json(json_string):
    """
    Parses JSON String and returns corresponding dictionary
    :param json_string:
    :return:
    """
    return json.loads(json_string)


def from_json_file(file_name):
    """
    Parses JSON from file and returns corresponding instance
    :param file_name: File containing JSON formatted object
    :return: A dictionary representing the JSON
    """
    return from_json(file_utils.read_entire_file_into_string(file_name))



import copy
from enum import Enum

from fjlc.utils.file_utils import read_entire_file_into_string
from fjlc.utils.json_utils import from_json

options = {}
intensifiers = {}
negators = set()
stop_words = set()


def is_stop_word(word):
    return word in stop_words


def is_negation(word):
    return word in negators


def is_intensifier(word):
    return word in intensifiers


def contains_stop_words(words):
    for word in words:
        if is_stop_word(word):
            return True

    return False


def contains_negation(words):
    for word in words:
        if is_negation(word):
            return True

    return False


def contains_intensifier(words):
    for word in words:
        if is_intensifier(word):
            return True

    return False


def get_variable(variable):
    return options[variable.name]


def set_variable(variable, value):
    options[variable.name] = value


def get_options():
    return copy.deepcopy(options)


def is_special_class_word(word):
    return word.startswith("||") and word.endswith("||")


def load_options(file_name):
    """
    Loads options from a JSON file. The file should contain general classifier options, intensifier words with their
    intensification values, negation words and stop words.

    @param file_name Name of file containing the options
    @throws IOException
    """
    words = from_json(read_entire_file_into_string(file_name))

    global options, intensifiers, negators, stop_words
    options = words["options"]
    intensifiers = words["intensifiers"]
    negators = words["negators"]
    stop_words = words["stopWords"]


def get_intensifier_value(word):
    intensifier = intensifiers.get(word, 0.0)
    mult = get_variable(Variable.AMPLIFIER_SCALAR) if intensifier > 0 else get_variable(
        Variable.DOWNTONER_SCALAR)
    return mult * intensifier


class Variable(Enum):
    NEGATION_VALUE = 0,
    EXCLAMATION_INTENSIFIER = 1,
    QUESTION_INTENSIFIER = 2,
    NEGATION_SCOPE_LENGTH = 3,
    DOWNTONER_SCALAR = 4,
    AMPLIFIER_SCALAR = 5,
    CLASSIFICATION_THRESHOLD_LOWER = 6,
    CLASSIFICATION_THRESHOLD_HIGHER = 7



"""
import com.freva.masteroppgave.lexicon.container.TokenTrie;
"""

import fjlc.classifier.classifier_options as classifier_options
from fjlc.classifier.sentence.lexical_token import LexicalToken
from fjlc.preprocessing.filters.regex_filters import RegexFilters


def lexically_parse_tweet(tweet, phrase_tree):
    """
    Returns list of LexicalTokens found in tweet. The list contains all the words in original tweet, but are
    optimally grouped up to form largest matching n-grams from lexicon. If no match is found, token is added as
    singleton.

    @param tweet      Tweet to lexically parse
    @param phrase_tree Token tree that contains all the lexical n-grams
    @return List of LexicalTokens
    """
    lexical_tokens = []

    prev = 0

    while True:
        match = RegexFilters.SENTENCE_END_PUNCTUATION.search(tweet[prev:])
        if match is None:
            break
        span = match.span()
        sentence = tweet[prev:prev + span[0]]
        punctuation = match.group(0)
        prev += span[1]
        lexical_tokens.extend(parse_sentence(sentence, punctuation, phrase_tree))

    lexical_tokens.extend(parse_sentence(tweet[prev:], None, phrase_tree))

    return lexical_tokens


def parse_sentence(sentence, punctuation, phrase_tree):
    sentence_tokens = RegexFilters.WHITESPACE.split(sentence)

    tokenized_sentence = phrase_tree.find_optimal_tokenization(sentence_tokens)
    tokens = list(map(lambda s: LexicalToken(s), tokenized_sentence))

    if len(tokens) > 0:
        tokens[len(tokens) - 1].set_at_end_of_sentence(True)

        if punctuation is not None and "!" in punctuation:
            exclamation_intensifier = classifier_options.get_variable(classifier_options.Variable.EXCLAMATION_INTENSIFIER)
            for token in tokens:
                token.intensify_token(exclamation_intensifier)

        elif punctuation is not None and "?" in punctuation:
            question_intensifier = classifier_options.get_variable(classifier_options.Variable.QUESTION_INTENSIFIER)
            for token in tokens:
                token.intensify_token(question_intensifier)

    return tokens



import html
import re
import fjlc.classifier.classifier_options as classifier_options
from fjlc.preprocessing.filters.regex_filters import RegexFilters
from normalizr import Normalizr

normalizr = Normalizr(language="en")


class Filters:
    USERNAME_PLACEHOLDER = " ||username|| "
    HASHTAG_PLACEHOLDER = " ||hashtag|| "
    RTTAG_PLACEHOLDER = " ||rt|| "
    URL_PLACEHOLDER = " ||url|| "

    def __init__(self, string_filters, token_filters):
        self.string_filters = string_filters
        self.token_filters = token_filters

    def apply(self, text):
        text = self.string_chain(text, self.string_filters)
        return self.token_chain(text, self.token_filters).strip()

    @staticmethod
    def string_chain(text, filters):
        """
        Chain several filters after each other, applies the filter on the entire string
        :param text: String to format
        :param filters: Sequence of filters to apply on String
        :return: The formatted String
        """
        if filters is None:
            return text

        for filter_function in filters:
            text = filter_function(text)

        return text

    @staticmethod
    def token_chain(text, filters):
        """
        Chain several filters after each other, applying filters only on non special class tokens as detected by
        {@link ClassifierOptions#isSpecialClassWord(String)}

        :param text: String to format
        :param filters: Sequence of filters to apply to tokens
        :return: The formatted String
        """
        if filters is None:
            return text

        sb = ""
        for token in RegexFilters.WHITESPACE.split(text):
            if not classifier_options.is_special_class_word(token):
                token = Filters.string_chain(token, filters)

            sb += token + " "

        return sb

    @staticmethod
    def html_unescape(text):
        """
        Returns HTML unescaped string

        :param text: String to format (f.ex. "&lt;3")
        :return: The formatted String (f.ex. "<3")
        """
        return html.unescape(text)

    @staticmethod
    def normalize_form(text):
        """
        Normalizes String to Latin characters if possible. WARNING: This also applies non-ASCII filter to the entire string

        :param text: String to format (f.ex. "A strîng wìth fúnny chäracters")
        :return: The formatted String (f.ex. "A string with funny characters")
        """
        return normalizr.remove_accent_marks(text)

    @staticmethod
    def remove_repeated_whitespace(text):
        """
        Removes repeated whitespace

        :param text: String to format (f.ex. "A string    with maany   spaces  ")
        :return: The formatted String (f.ex. "A string with many spaces ")
        """
        return RegexFilters.replace_whitespace(text, " ")

    @staticmethod
    def parse_unicode_emojis_to_alias(text):
        """
        Uses {@link EmojiParser#parseFromUnicode(String, EmojiParser.EmojiTransformer)} to parse unicode emojis to ASCII
        string " ||emoji_alias|| ".

        :param text: String to format (f.ex. "Hey \uD83D\uDC66\uD83C\uDFFF!")
        :return: The formatted String (f.ex. "Hey  ||boy|| !")
        """
        #return EmojiParser.parseFromUnicode(text, EMOJI_ALIAS_TRANSFORMER)
        return text

    @staticmethod
    def remove_unicode_emoticons(text):
        return normalizr.replace_emojis(text)

    @staticmethod
    def parse_emoticons(text):
        return RegexFilters.replace_emoticons(text, " ||$1|| ")

    @staticmethod
    def remove_emoticons(text):
        return RegexFilters.replace_emoticons(text, "")

    @staticmethod
    def remove_username(text):
        return RegexFilters.replace_username(text, "")

    @staticmethod
    def placeholder_username(text):
        return RegexFilters.replace_username(text, Filters.USERNAME_PLACEHOLDER)

    @staticmethod
    def remove_email(text):
        return re.sub(RegexFilters.TWITTER_EMAIL, "", text)

    @staticmethod
    def remove_hashtag(text):
        return RegexFilters.replace_hashtag(text, "")

    @staticmethod
    def placeholder_hashtag(text):
        return RegexFilters.replace_hashtag(text, Filters.HASHTAG_PLACEHOLDER)

    @staticmethod
    def hashtag_to_word(text):
        return RegexFilters.replace_hashtag(text, "$1")

    @staticmethod
    def protect_hashtag(text):
        return RegexFilters.replace_hashtag(text, " ||#$1|| ")

    @staticmethod
    def remove_rt_tag(text):
        return RegexFilters.replace_rt_tag(text, "")

    @staticmethod
    def placeholder_rt_tag(text):
        return RegexFilters.replace_rt_tag(text, Filters.RTTAG_PLACEHOLDER)

    @staticmethod
    def remove_url(text):
        return RegexFilters.replace_url(text, "")

    @staticmethod
    def placeholder_url(text):
        return RegexFilters.replace_url(text, Filters.URL_PLACEHOLDER)

    @staticmethod
    def remove_inner_word_characters(text):
        """
        Removes characters which are often part of a word (mostly apostrophes)

        :param text: String to format (f.ex. "Here's a sentence!")
        :return: The formatted String (f.ex. "Heres a sentence!")
        """
        return RegexFilters.replace_inner_word_characters(text, "")

    @staticmethod
    def remove_non_syntactical_text(text):
        """
        Removes all non-alphabetic or basic punctuation characters (!?,. )

        :param text: String to format (f.ex. "This is' a #crazy tæst")
        :return: The formatted String (f.ex. "This is a crazy tst")
        """
        return RegexFilters.replace_non_syntactical_text(text, " ")

    @staticmethod
    def remove_non_syntactical_text_plus(text):
        return RegexFilters.replace_non_syntactical_text_plus(text, " ")

    @staticmethod
    def remove_non_alphanumerical_text(text):
        """
        Removes non-alphanumerical characters

        :param text: String to format (f.ex "It's very nice!")
        :return: The formatted String (f.ex "It s very nice ")
        """
        return RegexFilters.replace_non_alphanumerical_text(text, " ")

    @staticmethod
    def remove_non_alphabetic_text(text):
        """
        Removes non alphabetic characters

        :param text: String to format (f.ex "Hey, m8!")
        :return: The formatted String (f.ex. "Hey m")
        """
        return RegexFilters.replace_non_alphabetic_text(text, "")

    @staticmethod
    def remove_free_digits(text):
        """
        Removes free standing digits (digits not part of a word)

        :param text: String to format (f.ex. "Only 90s kids will get this 1337 m8")
        :return: The formatted String (f.ex. "Only 90s kids will get this m8")
        """
        return RegexFilters.replace_free_digits(text, " ")

    @staticmethod
    def remove_repeating_characters(text):
        """
        Replaces repeating characters in String

        :param text: String to format (f.ex. "Today is a greeeeeeaaaaaaat dayy!")
        :return: The formatted String (f.ex. "Today is a great day!")
        """
        return RegexFilters.replace_repeating_characters(text, "$1")



import fjlc.classifier.classifier_options as classifier_options


class LexicalToken:
    def __init__(self, phrase):
        self.phrase = phrase
        self.lexical_value = 0.0
        self.intensification = 1

        self.in_negated_context = False
        self.at_end_of_sentence = False

    def get_phrase(self):
        return self.phrase

    def set_lexical_value(self, lexical_value):
        self.lexical_value = lexical_value

    def get_sentiment_value(self):
        sentiment_value = self.lexical_value
        if self.is_under_intensification():
            sentiment_value *= self.intensification

        if self.is_in_negated_context() and sentiment_value != 0:
            negation_value = classifier_options.get_variable(classifier_options.Variable.NEGATION_VALUE)
            sentiment_value = sentiment_value - negation_value if sentiment_value > 0 else sentiment_value + negation_value

        return sentiment_value

    def set_in_negated_context(self, in_negated_context):
        self.in_negated_context = in_negated_context

    def is_in_negated_context(self):
        return self.in_negated_context

    def set_at_end_of_sentence(self, at_end_of_sentence):
        self.at_end_of_sentence = at_end_of_sentence

    def is_at_the_end_of_sentence(self):
        return self.at_end_of_sentence

    def intensify_token(self, intensification):
        self.intensification = intensification

    def is_under_intensification(self):
        return self.intensification != 1

    def to_string(self):
        return "[" + self.phrase + (
            "_NEG" if self.is_in_negated_context() else "") + " | " + self.get_sentiment_value() + " | " + str(
            self.intensification) + "]"



def count_lines(file_name):
    try:
        f = open(file_name)
        count = sum(1 for line in f)
        f.close()
        return count
    except IOError:
        raise IOError


def read_entire_file_into_string(file_name):
    with open(file_name) as f:
        return f.read()


def write_to_file(file, data):
    with open(file) as f:
        f.write(data)



from os import path

import fjlc.classifier.classifier_options as classifier_options
import fjlc.lexical_classifier as lexical_classifier
import fjlc.preprocessing.filters.canonical_form as canonical_form
from fjlc.classifier.classifier import Classifier
from fjlc.lexicon.container.prior_polarity_lexicon import PriorPolarityLexicon
from fjlc.preprocessing.filters.filters import Filters
from fjlc.preprocessing.preprocessors.tweet_n_grams_pmi import TweetNGramsPMI
from fjlc.utils import json_utils, map_utils
from fjlc.lexicon.lexicon_creator import LexiconCreator
from fjlc.utils.reader.data_set_reader import DataSetReader
from fjlc.utils.reader.line_reader import LineReader

N_GRAM_STRING_FILTERS = [
    Filters.html_unescape, Filters.remove_unicode_emoticons, Filters.normalize_form, Filters.remove_url,
    Filters.remove_rt_tag, Filters.remove_hashtag, Filters.remove_username, Filters.remove_emoticons,
    Filters.remove_free_digits, str.lower
]
N_GRAM_CHARACTER_FILTERS = [
    Filters.remove_inner_word_characters, Filters.remove_non_syntactical_text, canonical_form.correct_word_via_canonical
]
N_GRAM_FILTERS = Filters(N_GRAM_STRING_FILTERS, N_GRAM_CHARACTER_FILTERS)

TWEET_STRING_FILTERS = [
    Filters.html_unescape, Filters.parse_unicode_emojis_to_alias, Filters.normalize_form, Filters.remove_url,
    Filters.remove_rt_tag, Filters.protect_hashtag, Filters.remove_email, Filters.remove_username,
    Filters.remove_free_digits, Filters.parse_emoticons, str.lower
]
TWEET_CHARACTER_FILTERS = [Filters.remove_inner_word_characters, Filters.remove_non_alphanumerical_text,
                           canonical_form.correct_word_via_canonical]
TWEET_FILTERS = Filters(TWEET_STRING_FILTERS, TWEET_CHARACTER_FILTERS)


class Lexicon:
    def __init__(self, n_grams_file, data_set_file, lexicon_file, max_error_rate, sentiment_value_threshold):
        self.n_grams_file = n_grams_file
        self.data_set_file = data_set_file
        self.lexicon_file = lexicon_file
        self.max_error_rate = max_error_rate
        self.sentiment_value_threshold = sentiment_value_threshold

    def create_lexicon(self):
        frequent_n_grams = json_utils.from_json_file(self.n_grams_file)
        data_set_reader = DataSetReader(self.data_set_file, 1, 0)

        lexicon_creator = LexiconCreator()
        # ProgressBar.track_progress(lexicon_creator, "Creating lexicon...")
        lexicon = lexicon_creator.create_lexicon(data_set_reader, frequent_n_grams, self.max_error_rate,
                                                 self.sentiment_value_threshold, TWEET_FILTERS)
        json_utils.to_json_file(self.lexicon_file, map_utils.sort_map_by_value(lexicon), True)

    @staticmethod
    def generate_n_grams(input_file, output_file, n_gram_range, cutoff_frequency, pmi_value_threshold):
        tweet_n_grams = TweetNGramsPMI()
        # ProgressBar.track_progress(tweet_n_grams, "Generating tweet n-grams...")
        ngrams = tweet_n_grams.get_frequent_n_grams(LineReader(input_file), n_gram_range, cutoff_frequency,
                                                    pmi_value_threshold, N_GRAM_FILTERS)

        json_utils.to_json_file(output_file, ngrams, True)


class LexiconClassifier:
    def __init__(self,
                 lexicon=path.join(path.abspath(path.dirname(__file__)), "res/data/lexicon.pmi.json"),
                 options=path.join(path.abspath(path.dirname(__file__)), "res/data/options.pmi.json"),
                 dictionary=path.join(path.abspath(path.dirname(__file__)), "res/dictionary.json")):
        self.lexicon = lexicon
        self.options = options
        self.dictionary = dictionary

        classifier_options.load_options(self.options)
        canonical_form.load_dictionary(self.dictionary)

        self.prior_polarity_lexicon = PriorPolarityLexicon(self.lexicon)
        self.classifier = Classifier(self.prior_polarity_lexicon, lexical_classifier.CLASSIFIER_FILTERS)

    def classify(self, tweets):
        """
        Classify tweet or tweets
        :param tweets: String or array of strings to classify.
        :return: String or array of strings depicting sentiment. Sentiment can be POSITIVE, NEGATIVE or NEUTRAL.
        """
        if type(tweets) == str:
            return self.classifier.classify(tweets)

        return list(map(lambda tweet: self.classifier.classify(tweet), tweets))

    def calculate_sentiment(self, tweets):
        """
        Classify tweet or tweets
        :param tweets: String or array of strings to classify.
        :return: Float or array of floats depicting sentiment value.
        """
        if type(tweets) == str:
            return self.classifier.calculate_sentiment(tweets)

        return list(map(lambda tweet: self.classifier.calculate_sentiment(tweet), tweets))



from enum import Enum
import re
from fjlc.utils.reader.line_reader import LineReader


TAB_REGEX = re.compile("\t")


class DataSetReader:

    def __init__(self, file_name, tweet_index, class_index):
        self.line_reader = LineReader(file_name)
        self.tweet_index = tweet_index
        self.class_index = class_index

    def has_next(self):
        return self.line_reader.has_next()

    def __next__(self):
        if not self.has_next():
            raise StopIteration
        return DataSetEntry(self.line_reader.__next__(), self.tweet_index, self.class_index)

    def __iter__(self):
        return self

    def get_progress(self):
        return self.line_reader.get_progress()


class DataSetEntry:

    def __init__(self, line, tweet_index, class_index):
        values = TAB_REGEX.split(line)
        self.tweet = values[tweet_index]
        self.classification = Classification.parse_classification_from_string(values[class_index])

    def get_tweet(self):
        return self.tweet

    def get_classification(self):
        return self.classification

    def apply_filters(self, filters):
        self.tweet = filters.apply(self.tweet)


class Classification(Enum):
    POSITIVE = 0
    NEUTRAL = 1
    NEGATIVE = 2

    @staticmethod
    def parse_classification_from_string(classification):
        if classification == "positive":
            return Classification.POSITIVE.name
        if classification == "neutral":
            return Classification.NEUTRAL.name
        if classification == "negative":
            return Classification.NEGATIVE.name

    @staticmethod
    def classify_from_thresholds(value, low_thresh, high_thresh):
        if value < low_thresh:
            return Classification.NEGATIVE.name
        elif value > high_thresh:
            return Classification.POSITIVE.name
        return Classification.NEUTRAL.name

    def is_positive(self):
        return self == Classification.POSITIVE

    def is_neutral(self):
        return self == Classification.NEUTRAL

    def is_negative(self):
        return self == Classification.NEGATIVE



def sort_map_by_value(dictionary):
    """
    Sorts Map by value. Map values must implement Comparable.
    
    :param dictionary: Map to sort 
    :return: Sorted map 
    """
    return sorted(dictionary, key=dictionary.get)


def normalize_map_between(dictionary, norm_min, norm_max):
    """
    Performs linear normalization of all values in Map between normMin and normMax

    :param: map     Map to normalize values for
    :param: normMin Smallest normalized value
    :param: normMax Largest normalized value
    :return: A new map with double values within [normMin, normMax]
    """
    if len(dictionary) < 2:
        return {}

    values = list(dictionary.values())

    norm_range = norm_max - norm_min
    map_min = min(values)
    map_range = max(values) - map_min
    range_factor = norm_range / float(map_range)

    normalized_map = {}
    for key, value in dictionary.items():
        normalized_map[key] = norm_min + (value - map_min) * range_factor

    return normalized_map

"""
    /**
     * Sorts map given a comparator
     *
     * @param map        Map to sort
     * @param comparator Comparator used to sort elements
     * @return Sorted map
     */
    public static <K, V> Map<K, V> sortMapWithComparator(Map<K, V> map, Comparator<Map.Entry<K, V>> comparator) {
        LinkedList<Map.Entry<K, V>> list = new LinkedList<>(map.entrySet());
        Collections.sort(list, comparator);

        Map<K, V> sortedHashMap = new LinkedHashMap<>();
        for (Map.Entry<K, V> entry : list) {
            sortedHashMap.put(entry.getKey(), entry.getValue());
        }

        return sortedHashMap;
    }


    /**
     * Increments value of key by increment if present in the list, otherwise initializes the value to increment.
     *
     * @param map       Map to increment key for
     * @param key       Key to increment
     * @param increment Value to increment by
     */
    public synchronized static <T> void incrementMapByValue(Map<T, Integer> map, T key, int increment) {
        map.put(key, map.getOrDefault(key, 0) + increment);
    }


    /**
     * Removes elements from map that are strictly smaller than the threshold element
     *
     * @param map    Map to remove items from
     * @param thresh Threshold element
     */
    public static <K, V extends Comparable<V>> void removeInfrequentItems(Map<K, V> map, V thresh) {
        Iterator<Map.Entry<K, V>> iter = map.entrySet().iterator();
        while (iter.hasNext()) {
            Map.Entry<K, V> entry = iter.next();
            if (entry.getValue().compareTo(thresh) < 0) {
                iter.remove();
            }
        }
    }

    /**
     * Extracts the map entries with key values found in the toExtract set
     *
     * @param map       Map to extract items from
     * @param toExtract The key values
     */
    public static <K, V> Map<K, V> extractItems(Map<K, V> map, Set<K> toExtract) {
        Map<K, V> extractedItems = new HashMap<>();
        toExtract.stream().filter(map::containsKey).forEach(key -> extractedItems.put(key, map.get(key)));
        return extractedItems;
    }

    /**
     * Merges two maps into a new map
     *
     * @param map1 Map to merge
     * @param map2 Map to merge
     */
    public static <K, V> Map<K, V> mergeMaps(Map<K, V> map1, Map<K, V> map2) {
        Map<K, V> mergedMap = new HashMap<>();
        mergedMap.putAll(map1);
        mergedMap.putAll(map2);
        return mergedMap;
    }
}
"""



import math

from fjlc.classifier import classifier_options
from fjlc.preprocessing.filters.regex_filters import RegexFilters


class TweetNGramsPMI:

    tweet_reader = None
    n_gram_tree = None

    def get_frequent_n_grams(self, input_reader, n, min_frequency, min_pmi, filters):
        """
        Finds all frequent (and meaningful) n-grams in a file, treating each new line as a new document.
        
        :param input_reader:    LineReader initialized on file with documents to generate n-grams for
        :param n:               Maximum n-gram length
        :param min_frequency:   Smallest required frequency to include n-gram
        :param min_pmi:         Minimum PMI value for n-gram to be included
        :param filters:         List of filters to apply to document before generating n-grams
        :return:                Map of n-grams as key and number of occurrences as value
        """
        line_counter = 0
        TweetNGramsPMI.tweet_reader = input_reader
        TweetNGramsPMI.n_gram_tree = self.NGramTree()

        # Todo: Parallelize
        for tweet in self.tweet_reader:
            line_counter += 1
            if line_counter % 200000 == 0:
                TweetNGramsPMI.n_gram_tree.prune_infrequent(math.ceil(min_frequency * line_counter / 2.))

            tweet = filters.apply(tweet)
            for sentence in RegexFilters.SENTENCE_END_PUNCTUATION.split(tweet):
                tokens = RegexFilters.WHITESPACE.split(sentence.strip())
                if len(tokens) == 1:
                    continue

                for i in range(len(tokens)):
                    self.n_gram_tree.increment_n_gram(tokens[i:min(i + n, len(tokens))])

        return self.n_gram_tree.get_n_grams(int(min_frequency * line_counter), min_pmi)

    @staticmethod
    def get_progress():
        return 0 if TweetNGramsPMI.tweet_reader is None else TweetNGramsPMI.tweet_reader.get_progress()

    class NGramTree:

        def __init__(self):
            self.root = TweetNGramsPMI.Node("")

        def increment_n_gram(self, n_gram):
            current = self.root
            current.num_occurrences += 1

            for word in n_gram:
                if not current.has_child(word):
                    current.add_child(word)

                current = current.get_child(word)
                current.num_occurrences += 1

        def get_node(self, phrase):
            current = self.root
            for word in RegexFilters.WHITESPACE.split(phrase):
                if not current.has_child(word):
                    return None

                current = current.get_child(word)
            return current

        def prune_infrequent(self, limit):
            self.root.prune_infrequent(limit)

        def get_n_grams(self, limit, inclusion_threshold):
            all_n_grams = {}

            for child in self.root.children.values():
                child.add_frequent_phrases(all_n_grams, limit, child.phrase)

            filtered_n_grams = []
            for next_key, next_value in all_n_grams.items():
                n_gram_tokens = RegexFilters.WHITESPACE.split(next_key)

                if next_value >= inclusion_threshold and not classifier_options.contains_intensifier(n_gram_tokens) and not classifier_options.is_stop_word(n_gram_tokens[-1]):
                    filtered_n_grams.append(next_key)

            return filtered_n_grams

    class Node:
        def __init__(self, phrase):
            self.children = {}
            self.phrase = phrase
            self.num_occurrences = 0
            self.log_score = 0.0

        def has_child(self, value):
            return value in self.children

        def add_child(self, value):
            self.children[value] = TweetNGramsPMI.Node(value)

        def get_child(self, value):
            return self.children[value]

        def get_log_score(self):
            if self.log_score == 0.0:
                self.log_score = math.log(self.num_occurrences)
            return self.log_score

        def prune_infrequent(self, limit):
            self.children = [child for child in self.children if child.get_value().num_occurrences >= limit]
            for child in self.children:
                child.get_value().prune_infrequent(limit)

        def add_frequent_phrases(self, dictionary, limit, prefix):
            for child in self.children.values():
                if child.num_occurrences < limit:
                    continue
                last_word = TweetNGramsPMI.n_gram_tree.get_node(child.phrase)

                if last_word is not None and last_word.num_occurrences >= limit:
                    temp = TweetNGramsPMI.n_gram_tree.root.get_log_score() + child.get_log_score() - self.get_log_score() - last_word.get_log_score()

                    candidate = prefix + " " + child.phrase
                    dictionary[candidate] = temp
                    child.add_frequent_phrases(dictionary, limit, candidate)



import fjlc.classifier.classifier_options as classifier_options
import fjlc.classifier.sentence.lexical_parser as lexical_parser
from fjlc.lexicon.container.token_trie import TokenTrie
from fjlc.utils.reader.data_set_reader import Classification


def propagate_negation(lexical_tokens, index):
    negation_scope_length = classifier_options.get_variable(classifier_options.Variable.NEGATION_SCOPE_LENGTH)
    for i in range(index + 1, int(min(index + negation_scope_length, len(lexical_tokens)))):
        lexical_tokens[i].set_in_negated_context(True)
        if lexical_tokens[i].is_at_the_end_of_sentence():
            break


def intensify_next(lexical_tokens, index, intensification):
    if not lexical_tokens[index].is_at_the_end_of_sentence():
        lexical_tokens[index + 1].intensify_token(intensification)


class Classifier:
    def __init__(self, lexicon, filters=None):
        self.lexicon = lexicon
        self.filters = filters
        self.phrase_tree = TokenTrie(lexicon.get_subjective_words())

    def classify(self, tweet):
        """
        Classifies the tweet into one of three classes (negative, neutral or positive) depending on the sentiment value
        of the tweet and the thresholds specified in the classifier_options

        :param tweet: String tweet to classify
        :return: Sentiment classification (negative, neutral or positive)
        """
        sentiment_value = self.calculate_sentiment(tweet)

        return Classification.classify_from_thresholds(sentiment_value,
                                                       classifier_options.get_variable(
                                                           classifier_options.Variable.CLASSIFICATION_THRESHOLD_LOWER),
                                                       classifier_options.get_variable(
                                                           classifier_options.Variable.CLASSIFICATION_THRESHOLD_HIGHER))

    def calculate_sentiment(self, tweet):
        if self.filters is not None:
            tweet = self.filters.apply(tweet)

        lexical_tokens = lexical_parser.lexically_parse_tweet(tweet, self.phrase_tree)
        for i in range(len(lexical_tokens)):
            token = lexical_tokens[i]
            phrase = token.get_phrase()

            if self.lexicon.has_token(phrase):
                token.set_lexical_value(self.lexicon.get_token_polarity(phrase))

            elif classifier_options.is_negation(phrase):
                propagate_negation(lexical_tokens, i)

            elif classifier_options.is_intensifier(phrase):
                intensify_next(lexical_tokens, i, classifier_options.get_intensifier_value(phrase))

        return sum(map(lambda t: t.get_sentiment_value(), lexical_tokens))



import re

# Emoticon definitions.
NormalEyes = "[:=8]"
WinkEyes = "[;]"
NoseArea = "[Ooc^*'-]?"
HappyMouths = "[)*>}\\]]"
SadMouths = "[c<|@L{/\\(\\[]"
conditionalEmotes = "([x=;:]" + NoseArea + "d|[=;:]p)"


class RegexFilters:
    EMOTICON_NEGATIVE = re.compile("(" + NormalEyes + NoseArea + SadMouths + ")")
    EMOTICON_POSITIVE = re.compile("(\\^_\\^|<3+|" + "((" + NormalEyes + "|" +
                                   WinkEyes + ")" + NoseArea + HappyMouths + "))")
    EMOTICON_CONDITIONAL_LEFT = re.compile("(?:(?:^|\\s)" + conditionalEmotes + ")", re.IGNORECASE)
    EMOTICON_CONDITIONAL_RIGHT = re.compile("(?:" + conditionalEmotes + "(?:$|\\s))", re.IGNORECASE)

    # Twitter basic elements
    TWITTER_USERNAME = re.compile("(@\\w{1,15})")
    TWITTER_HASHTAG = re.compile("#([a-zA-Z]+\\w*)")
    TWITTER_RT_TAG = re.compile("(^RT\\s+|\\s+RT\\s+)")
    TWITTER_URL = re.compile("((https?://|www)\\S+)")
    TWITTER_EMAIL = re.compile("\\w+@\\S+")

    WHITESPACE = re.compile("\\s+")
    INNER_WORD_CHAR = re.compile("['`´’]")
    NON_SYNTACTICAL_TEXT = re.compile("[^a-z ?!.,]", re.IGNORECASE)
    NON_SYNTACTICAL_TEXT_PLUS = re.compile("[^a-z ?!.]", re.IGNORECASE)
    SENTENCE_END_PUNCTUATION = re.compile("[!?,.]")

    NON_ALPHANUMERIC_TEXT = re.compile("[^a-zA-Z0-9 ]")
    NON_ALPHABETIC_TEXT = re.compile("[^a-zA-Z ]")
    NON_ASCII_CHARACTERS = re.compile("[^\\p{ASCII}]")
    FREE_DIGITS = re.compile("([^\\w]|^)[0-9]+([^\\w]+[0-9]+)*([^\\w]|$)")
    REPEATING_CHARACTERS = re.compile("(.)\\1+")

    @staticmethod
    def replace_emoticons(text, replace):
        new_text = re.sub(RegexFilters.EMOTICON_POSITIVE, replace, text)
        new_text = re.sub(RegexFilters.EMOTICON_NEGATIVE, replace, new_text)
        new_text = re.sub(RegexFilters.EMOTICON_CONDITIONAL_LEFT, replace, new_text)
        return re.sub(RegexFilters.EMOTICON_CONDITIONAL_RIGHT, replace, new_text)

    @staticmethod
    def replace_username(text, replace):
        return re.sub(RegexFilters.TWITTER_USERNAME, replace, text)

    @staticmethod
    def replace_hashtag(text, replace):
        return re.sub(RegexFilters.TWITTER_HASHTAG, replace, text)

    @staticmethod
    def replace_rt_tag(text, replace):
        return re.sub(RegexFilters.TWITTER_RT_TAG, replace, text)

    @staticmethod
    def replace_url(text, replace):
        return re.sub(RegexFilters.TWITTER_URL, replace, text)

    @staticmethod
    def replace_whitespace(text, replace):
        return re.sub(RegexFilters.WHITESPACE, replace, text)

    @staticmethod
    def replace_inner_word_characters(text, replace):
        return re.sub(RegexFilters.INNER_WORD_CHAR, replace, text)

    @staticmethod
    def replace_non_syntactical_text(text, replace):
        return re.sub(RegexFilters.NON_SYNTACTICAL_TEXT, replace, text)

    @staticmethod
    def replace_non_syntactical_text_plus(text, replace):
        return re.sub(RegexFilters.NON_SYNTACTICAL_TEXT_PLUS, replace, text)

    @staticmethod
    def replace_non_alphanumerical_text(text, replace):
        return re.sub(RegexFilters.NON_ALPHANUMERIC_TEXT, replace, text)

    @staticmethod
    def replace_non_alphabetic_text(text, replace):
        return re.sub(RegexFilters.NON_ALPHABETIC_TEXT, replace, text)

    @staticmethod
    def replace_free_digits(text, replace):
        return re.sub(RegexFilters.FREE_DIGITS, replace, text)

    @staticmethod
    def replace_non_ascii(text, replace):
        return re.sub(RegexFilters.NON_ASCII_CHARACTERS, replace, text)

    @staticmethod
    def replace_repeating_characters(text, replace):
        return re.sub(RegexFilters.REPEATING_CHARACTERS, replace, text)



import sys
import fjlc.utils.file_utils as file_utils
import fjlc.utils.json_utils as json_utils
from fjlc.preprocessing.filters.filters import Filters

dictionary = {}


def load_dictionary(file_name):
    global dictionary
    dictionary = json_utils.from_json(file_utils.read_entire_file_into_string(file_name))


def correct_word_via_canonical(text):
    canonical = Filters.remove_repeating_characters(text)

    candidates = dictionary.get(canonical, None)
    if candidates is None:
        return text
    elif len(candidates) == 1:
        return candidates[0]

    closest_dist = sys.maxsize
    closest_string = canonical
    for candidate in candidates:
        dist = levenshtein_distance(text, candidate)
        if dist < closest_dist:
            closest_dist = dist
            closest_string = candidate

    return closest_string


def levenshtein_distance(s1, s2):
    """
    Calculates edit distance between two strings without replacement

    :param s1: String one
    :param s2: String two
    :return: Minimum number of insertions/deletions between the two strings to make them equal
    """
    if len(s1) < len(s2):
        return levenshtein_distance(s2, s1)

    # len(s1) >= len(s2)
    if len(s2) == 0:
        return len(s1)

    previous_row = range(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[
                             j + 1] + 1  # j+1 instead of j since previous_row and current_row are one character longer
            deletions = current_row[j] + 1  # than s2
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]



from fjlc.preprocessing.filters.filters import Filters
from fjlc.preprocessing.filters.canonical_form import correct_word_via_canonical

CLASSIFIER_STRING_FILTERS = [
    Filters.html_unescape,
    Filters.parse_unicode_emojis_to_alias,
    Filters.normalize_form,
    Filters.remove_url,
    Filters.remove_rt_tag,
    Filters.protect_hashtag,
    Filters.remove_email,
    Filters.remove_username,
    Filters.parse_emoticons,
    Filters.remove_free_digits,
    str.lower
]

CLASSIFIER_CHARACTER_FILTERS = [
    Filters.remove_inner_word_characters,
    Filters.remove_non_syntactical_text,
    correct_word_via_canonical,
]

CLASSIFIER_FILTERS = Filters(CLASSIFIER_STRING_FILTERS, CLASSIFIER_CHARACTER_FILTERS)

TEST_SETS = {
    "2013-TEST": "res/semeval/2013-2-test-gold-B.tsv",
    "2014-TEST": "res/semeval/2014-9-test-gold-B.tsv",
    "2015-TEST": "res/semeval/2015-10-test-gold-B.tsv",
    "2016-TEST": "res/semeval/2016-4-test-gold-A.tsv"
}

"""
    public static void main(String[] args) throws IOException {
        long startTime = System.currentTimeMillis();
        ClassifierOptions.loadOptions(new File("res/data/options.pmi.json"));

        PriorPolarityLexicon priorPolarityLexicon = new PriorPolarityLexicon(new File("res/data/lexicon.pmi.json"));
        Classifier classifier = new Classifier(priorPolarityLexicon, CLASSIFIER_FILTERS);

        ClassificationCollection classificationCollection = new ClassificationCollection(Classification.values());
        for (Map.Entry<String, File> testSet : TEST_SETS.entrySet()) {
            Parallel.For(new DataSetReader(testSet.getValue(), 3, 2), entry -> {
                Classification predicted = classifier.classify(entry.getTweet());
                classificationCollection.updateEvidence(testSet.getKey(), entry.getClassification(), predicted);
            });
        }

        System.out.println(classificationCollection.getShortClassificationReport());
        System.out.println("In: " + (System.currentTimeMillis() - startTime) + "ms");
    }
}

"""



VOWELS = "aeiouy"
SEPARATOR = ","


def get_adverb_and_adjectives(word):
    if not consists_only_of_alphabetical_characters(word):
        return []

    adjectives = get_comparative_and_superlative_adjectives(word)
    if len(adjectives) == 0:
        return [form_adverb_from_adjective(word)]
    else:
        adjectives = form_adverb_from_adjective(word) + SEPARATOR + adjectives
        return adjectives.split(SEPARATOR)


def get_comparative_and_superlative_adjectives(word):
    """
    Returns a comma separated string with the comparative and the superlative forms of the input adjective. F.ex.
    "good" => "better,best", "happy" => "happier,happiest". If the comparative and superlative forms are
    "more [word]" and "most [word]", empty string is returned, f.ex. "careful" => "".
    Forms are generated using rules from: http://www.eflnet.com/tutorials/adjcompsup.php

    :param word: adjective
    :return: comma separated comparative and the superlative forms of the input adjective
    """
    if word == "good":
        return "better,best"
    if word == "bad":
        return "worse,worst"
    if word == "far":
        return "farther,farthest"
    if word == "little":
        return "less,least"
    if word == "slow":
        return "slower,slowest"

    return normal_comparative_and_superlative_adjectives(word)


def normal_comparative_and_superlative_adjectives(word):
    number_of_syllables = get_number_of_syllables(word)
    sb = ""

    if number_of_syllables == 1: # If one-syllable adjective
        last_letter = word[-1]

        # If the adjective ends with an e, just add –r for the comparative form and –st for the superlative form
        if word.endswith("e"):
            sb += word + "r"
            sb += SEPARATOR + word + "st"

        # If the adjective ends with –y, change the y to i and add –er for the comparative form.
        # For the superlative form change the y to i and add –est.
        elif word.endswith("y"):
            stub = word[:-1]
            sb += stub + "ier"
            sb += SEPARATOR + stub + "iest"

        # If the adjective ends with a single consonant with a vowel before it, double the consonant and add –er
        # for the comparative form; and double the consonant and add –est for the superlative form
        elif is_vowel(word[-2]) and not is_vowel(last_letter):
            sb += word + last_letter + "er"
            sb += SEPARATOR + word + last_letter + "est"

        # Otherwise just add -er for the comparative form and -est for the superlative form
        else:
            sb += word + "er"
            sb += SEPARATOR + word + "est"

    elif number_of_syllables == 2: # If two-syllable adjective
        # If the adjective ends with –y, change the y to i and add –er for the comparative form.
        # For the superlative form change the y to i and add –est.
        if word.endswith("y"):
            stub = word[:-1]
            sb += stub + "ier"
            sb += SEPARATOR + stub + "iest"

        # If the adjective ending in –er, -le, or –ow, add –er and –est to form the comparative and superlative forms
        elif word.endswith("er") or word.endswith("le") or word.endswith("ow"):
            sb += word + "er"
            sb += SEPARATOR + word + "est"

    return sb


def form_adverb_from_adjective(adjective):
    """
    Forms an adverb from the input adjective, f.ex. "happy" => "happily".
    Adverbs are generated using rules from: http://www.edufind.com/english-grammar/forming-adverbs-adjectives/

    :param adjective: adjective
    :return: adverb form of the input adjective
    """

    # If the adjective ends in -able, -ible, or -le, replace the -e with -y
    if adjective.endswith("able") or adjective.endswith("ible") or adjective.endswith("le"):
        return adjective[:-1] + "y"

    # If the adjective ends in -y, replace the y with i and add -ly
    elif adjective.endswith("y"):
        return adjective[:-1] + "ily"

    # If the adjective ends in -ic, add -ally
    elif adjective.endswith("ic"):
        return adjective[:-2] + "ally"

    # In most cases, an adverb is formed by adding -ly to an adjective
    return adjective + "ly"


def get_number_of_syllables(word):
    count = 0
    last_is_consonant = True

    for i in range(len(word) - 1):
        if is_vowel(word[i]):
            if last_is_consonant:
                count += 1
            last_is_consonant = False
        else:
            last_is_consonant = True

    return count


def is_vowel(character):
    return character in VOWELS


def consists_only_of_alphabetical_characters(name):
    return name.isalpha()



import math

from fjlc.lexicon.container.token_trie import TokenTrie
from fjlc.preprocessing.filters.regex_filters import RegexFilters
import fjlc.classifier.classifier_options as classifier_options
import fjlc.lexicon.container.adjectives as adjectives
import fjlc.utils.map_utils as map_utils


class LexiconCreator:

    def __init__(self):
        self.data_set_reader = None

    def create_lexicon(self, data_set_reader, n_grams, min_total_occurrences, min_sentiment_value, filters):
        """
         Generates sentiment lexicon using PMI on words and classification of context they are in.
         
         :param: dataSetReader       Dataset containing tweets and their sentiment classification
         :param: nGrams              n-grams to calculate sentiment for (with n>1, singletons are calculated automatically)
         :param: minTotalOccurrences minimum number of times n-gram must have appeared in dataset before a sentiment value
                                     is assigned (higher value gives more accurate sentiment value)
         :param: minSentimentValue   minimum sentiment value required to be included in lexicon (values close to 0 are
                                     often words that are used equally in positive or negative context, possibly even
                                     different words, but with same spelling, and thus having uncertain value)
         :param: filters             filters to apply to tweets before searching for n-grams
         :return: map of n-grams and their sentiment values, sentiment values are in [-5, 5]
        """
        counter = self.count_n_grams_py_polarity(data_set_reader, n_grams, filters)
        lexicon = {}

        pos = sum(map(lambda i: i.num_positive, counter.values()))
        neg = sum(map(lambda i: i.num_negative, counter.values()))
        ratio = neg / float(pos)

        for key, value in counter.items():
            if value.get_total_occurrences() <= min_total_occurrences:
                continue

            over = value.num_positive
            under = value.num_negative

            sentiment_value = math.log(ratio * over / under)
            if abs(sentiment_value) >= min_sentiment_value:
                lexicon[key] = sentiment_value

                if RegexFilters.WHITESPACE.split(key).length == 1 and not classifier_options.is_special_class_word(key):
                    for related_word in adjectives.get_adverb_and_adjectives(key):
                        if related_word in counter and related_word not in lexicon:
                            lexicon[related_word] = sentiment_value

        return map_utils.normalize_map_between(lexicon, -5, 5)

    def count_n_grams_py_polarity(self, data_set_reader, n_grams, filters):
        """
        Returns a map of n-gram and the number of times it appeared in positive context and the number of times it
        appeared in negative context in dataset file.

        :param data_set_reader: Dataset containing tweets and their classification
        :param n_grams: n-grams to count occurrences for
        :param filters: filters to apply to tweets in dataset before searching for n-grams
        :return: Map of Counter instances for n-grams in nGrams Collection
        """
        self.data_set_reader = data_set_reader
        token_trie = TokenTrie(n_grams)

        counter = {}

        # Todo: parallelize
        for entry in data_set_reader.items():
            tweet = filters.apply(entry.get_tweet())
            tokens = token_trie.find_optimal_tokenization(RegexFilters.WHITESPACE.split(tweet))

            for n_gram in tokens:
                n_gram_words = RegexFilters.WHITESPACE.split(n_gram)
                if self.contains_illegal_word(n_gram_words):
                    continue
                if not n_gram in counter:
                    counter[n_gram] = self.Counter()

                if entry.get_classification().is_positive():
                    counter[n_gram].num_positive += 1
                elif entry.get_classification().is_negative():
                    counter[n_gram].num_negative += 1

        return counter

    @staticmethod
    def contains_illegal_word(n_gram):
        return classifier_options.is_stop_word(n_gram[-1] or classifier_options.contains_intensifier(n_gram))

    def get_progress(self):
        return 0 if self.data_set_reader is None else self.data_set_reader.get_progress()

    class Counter:
        def __init__(self):
            self.num_positive = 4
            self.num_negative = 4

        def get_total_occurrences(self):
            return self.num_positive + self.num_negative



from fjlc.preprocessing.filters.regex_filters import RegexFilters
import fjlc.classifier.classifier_options as classifier_options
import functools


class TokenTrie:

    def __init__(self, sentences):
        """
        Creates a phrase trie for efficient sub-phrase look up

        @param sentences List of Strings of all the phrases which are whitespace delimited n-grams
        """
        self.root = TokenTrie.Node()
        for sentence in sentences:
            words = RegexFilters.WHITESPACE.split(sentence)
            self.add_token_sequence(words)

    def add_token_sequence(self, token_sequence):
        tree = self.root
        for token in token_sequence:
            if not tree.has_child(token):
                tree.add_child(token)
            tree = tree.get_child(token)
        tree.set_phrase_end(True)

    def has_tokens(self, phrase):
        """
        Checks if phrase or sub-phrase exists in the tree.

        If set of phrases contains phrases such as: "state", "of the" and "state of the art", look up on:
        "state" returns true, "of" returns null, "of the art" returns false.

        :param phrase: Phrase or sub-phrase to look up.
        :type: phrase: list of str
        :return: Returns true if phrase in its entirety is in the tree,
        null if part of the phrase matches a larger tokenSequence,
        false if phrases matches no other phrase entirely and not part any longer phrase.
        """

        if len(phrase) == 1 and classifier_options.is_special_class_word(phrase[0]):
            return True

        tree = self.root
        for token in phrase:
            if not tree.has_child(token):
                return False
            tree = tree.get_child(token)

        return True if tree.is_end_of_phrase() else None

    def find_tracked_words(self, tokens):
        """
        Finds word-ranges all of phrases in tokens stored in TokenTrie

        :param tokens: Sequence of tokens to find phrases in
        :type tokens: list of str
        :return: List of Tokens found in tokens
        """
        tracked_words = []

        for i in range(len(tokens)):
            for j in range(i + 1, len(tokens) + 1):
                phrase = tokens[i:j]
                status = self.has_tokens(phrase)

                if status is not None:
                    if status is True:
                        tracked_words.append(TokenTrie.Token(phrase, i, j - 1))
                    elif status is False:
                        break

        return tracked_words

    def find_optimal_allocation(self, tokens):
        """
        Finds longest, non-overlapping word-ranges of phrases in tokens stored in TokenTrie

        :param tokens: tokens tokenize
        :type tokens: list of str
        :return: Optimal allocation of tokens to phrases
        :rtype: list of TokenTrie.Token
        """
        token_ranges = self.find_tracked_words(tokens)
        token_ranges.sort()

        for offset in range(1, len(token_ranges)):
            to_be_removed = []
            for candidate in token_ranges[offset:]:
                for i in range(offset):
                    if token_ranges[i].overlaps_with(candidate):
                        to_be_removed.append(candidate)
                        break

            token_ranges = [token for token in token_ranges if token not in to_be_removed]

        token_ranges.sort(key=lambda token: token.get_start_index())
        return token_ranges

    def find_optimal_tokenization(self, tokens):
        """
        Similar to {@link #findOptimalAllocation(String[])}, but also includes the words not matching any longer n-gram
        in TokenTrie as singletons.

        :param tokens: tokens to tokenize
        :return: Optimal allocation of tokens to phrases, with non matching tokens as singletons.
        """
        token_ranges = self.find_optimal_allocation(tokens)
        tokenized_sentence = []

        set_index = 0
        for token in token_ranges:
            while set_index < token.get_start_index():
                tokenized_sentence.append(tokens[set_index])
                set_index += 1
            tokenized_sentence.append(" ".join(token.get_token_sequence()))
            set_index = token.get_end_index() + 1

        while set_index < len(tokens):
            tokenized_sentence.append(tokens[set_index])
            set_index += 1

        return tokenized_sentence

    @functools.total_ordering
    class Token:

        def __init__(self, token_sequence, start_index, end_index):
            self.token_sequence = token_sequence
            self.start_index = start_index
            self.end_index = end_index

        def get_token_sequence(self):
            return self.token_sequence

        def get_start_index(self):
            return self.start_index

        def get_end_index(self):
            return self.end_index

        def get_phrase_length(self):
            return self.end_index - self.start_index

        def overlaps_with(self, other):
            """
            Checks if two phrases overlap
            :param other: The other token_sequence
            :return: True if overlap, False otherwise
            """
            return self.get_start_index() <= other.get_end_index() and other.get_start_index() <= self.get_end_index()

        def __lt__(self, other):
            size_diff = other.get_phrase_length() - self.get_phrase_length()
            diff = size_diff if size_diff != 0 else other.get_start_index() - self.get_start_index()
            return diff < 0

        def __eq__(self, other):
            size_diff = other.get_phrase_length() - self.get_phrase_length()
            diff = size_diff if size_diff != 0 else other.get_start_index() - self.get_start_index()
            return diff == 0

        def __str__(self):
            return str(self.token_sequence)

        def __repr__(self):
            return "Token(" + str(self.start_index) + "," + str(self.end_index) + ")<" + str(self.token_sequence) + ">"

    class Node:

        def __init__(self):
            self.children = {}
            self.end_of_phrase = False

        def has_child(self, value):
            return value in self.children

        def add_child(self, value):
            self.children[value] = TokenTrie.Node()

        def get_child(self, value):
            return self.children[value]

        def set_phrase_end(self, phrase_end):
            self.end_of_phrase = phrase_end

        def is_end_of_phrase(self):
            return self.end_of_phrase



from fjlc.utils import file_utils


class LineReader:

    def __init__(self, file_name):
        self.total_lines = file_utils.count_lines(file_name)
        self.file = open(file_name)
        self.line_counter = 0

    def has_next(self):
        return self.line_counter < self.total_lines

    def __next__(self):
        if not self.has_next():
            raise StopIteration
        self.line_counter += 1
        return self.file.readline()

    def __iter__(self):
        return self

    def get_progress(self):
        return 0 if self.total_lines else 100.0 * self.line_counter / self.total_lines
